/* $Copyright$
 *  Copyright (C) Fujitsu Siemens Computers GmbH 2001
 *  All rights reserved
 */
#ident "$Header$"
#ifndef WITHOUT_INCLUDES
#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE_SUB
#  define MODULE
#endif
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/ioctl.h>
#include <linux/sched.h>
#include <linux/tqueue.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <asm/system.h>
#endif
#ifndef WITHOUT_INCLUDES
#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif
#if LINUX_VERSION_CODE >= 0x020400
static DECLARE_WAIT_QUEUE_HEAD(WaitQ);
#else
#define vm_pgoff                        vm_offset
#define pci_resource_start(dev, i)      (dev->base_address[i])
static struct wait_queue *WaitQ = NULL;
#endif
#endif
#ident "$Header: $"
#ident "$Header$"
#ident "$Header: $"
EXPORT_NO_SYMBOLS;
static const int SUCCESS = 0;
static const char *COPA_DEV = "copa";
static unsigned long copaRomStart;
static u32 pDoorbellAddress;
static unsigned char *pvirtRomStart;
static unsigned int *pProperties;
static u32 Properties;
static unsigned int *pDoorbellRegister=NULL;
static const unsigned long COPA_ROM_LEN = 0x4000;
static const unsigned int SNI_VENDOR_ID = 0x110A;
static const unsigned int COPA_F1_DEVICE_ID = 0x007C;
static const unsigned int COPA_F0_DEVICE_ID = 0x007B;
static const u8 COPA_DEVICE_CONTROL_REGISTER = 0x04;
static const u8 COPA_BASE_ADDRES_REGISTER0 = 0x10;
static const u8 COPA_BASE_ADDRES_REGISTER1 = 0x14;
static const u8 COPA_BASE_ADDRES_REGISTER2 = 0x18;
static const u8 COPA_INTERRUPT_LINE_REGISTER = 0x3C;
static const u8 COPA_INTERRUPT_PIN_REGISTER = 0x3D;
static const u8 COPA_MAILBOX_DATA = 0x40;
static const u8 COPA_MAILBOX_STATUS = 0x44;
static const u8 COPA_INTERRUPT_SET_REGISTER = 0x30;
static const u8 COPA_INTERRUPT_CLEAR_REGISTER = 0x34;
static short copaDebug = 0;
MODULE_PARM(copaDebug, "h");
static int deviceOpen = 0;
static int copa_major = 0;
static struct pci_dev *copaf1Dev = NULL;
static struct pci_dev *copaf0Dev = NULL;
static int copa_interrupt_received=0;
static int copa_open(struct inode *inode, struct file *file)
{
    if (copaDebug) printk(KERN_DEBUG "copa :" "copa_open: (%p,%p)\n",inode, file);
    if(deviceOpen) {
        if (copaDebug) printk(KERN_DEBUG "copa :" "copa_open: device is busy\n");
        return -EBUSY;
    }
    deviceOpen++;
    MOD_INC_USE_COUNT;
    return SUCCESS;
}
static int copa_release(struct inode *inode, struct file *file)
{
    deviceOpen--;
    if (copaDebug) printk(KERN_DEBUG "copa :" "copa_release: (%p,%p)\n",inode, file);
    MOD_DEC_USE_COUNT;
    return SUCCESS;
}
static void copa_vma_open(struct vm_area_struct *area) {
    if (copaDebug) printk(KERN_DEBUG "copa :" "copa_vma_open\n");
    MOD_INC_USE_COUNT;
}
static void copa_vma_close(struct vm_area_struct *area) {
    if (copaDebug) printk(KERN_DEBUG "copa :" "copa_vma_close\n");
    MOD_DEC_USE_COUNT;
}
static struct vm_operations_struct copa_vm_ops = {
    copa_vma_open,
    copa_vma_close,
};
static int copa_mmap(struct file *file, struct vm_area_struct *vma)
{
    int err;
    unsigned long virt_add = vma->vm_start;
    unsigned long vsize = vma->vm_end-vma->vm_start;
    if (copaDebug) printk(KERN_DEBUG "copa :" "copa_mmap virt=0x%lx phys=0x%lx(0x%lx) size=0x%lx(0x%lx)\n",virt_add, copaRomStart, (long unsigned int)vma->vm_pgoff, COPA_ROM_LEN, vsize);
    if (vsize > COPA_ROM_LEN) {
        if (copaDebug) printk(KERN_DEBUG "copa :" "copa_mmap: vsize > COPA_ROM_LEN\n");
        return -EINVAL;
    }
    err = remap_page_range(
            virt_add,
            copaRomStart,
            vsize,
            vma->vm_page_prot);
    if (err) {
        if (copaDebug) printk(KERN_DEBUG "copa :" "copa_mmap: remap_page_range failed: %d\n",err);
        return -EAGAIN;
    }
    if (vma->vm_ops) {
        if (copaDebug) printk(KERN_DEBUG "copa :" "copa_mmap: vma->vm_ops != NULL\n");
        return -EINVAL;
    }
    vma->vm_ops = &copa_vm_ops;
    MOD_INC_USE_COUNT;
    return SUCCESS;
}
static void copa_irq_handler(int irq,void *dev_id,struct pt_regs *regs)
{
        if (pDoorbellRegister)
                if (copaDebug) printk(KERN_DEBUG "copa :" "copa_irq_handler status 0x%08x\n",*pDoorbellRegister);
        if ((pDoorbellRegister) && (*pDoorbellRegister & (1<<24)))
        {
                *pDoorbellRegister |= (1<<24);
                if (copaDebug) printk(KERN_DEBUG "copa :" "Interrupt asserted\n");
        }
    wake_up(&WaitQ);
        copa_interrupt_received=1;
}
static int copa_ioctl(
                                struct inode *inode,
                                struct file *file,
                                unsigned int cmd,
                                unsigned long arg )
{
        long flags;
        if (copaDebug) printk(KERN_DEBUG "copa :" "-->copa_ioctl(0x%x)\n",cmd);
        if (_IOC_TYPE(cmd) != 'r') {
                if (copaDebug) printk(KERN_DEBUG "copa :" "  type != magic\n");
                return -EINVAL;
        }
        if (_IOC_NR(cmd) > 5) {
                if (copaDebug) printk(KERN_DEBUG "copa :" "  nr > maxnr\n");
                return -EINVAL;
        }
        switch(cmd)
        {
                case _IO('r', 0):
                        save_flags(flags);
                        cli();
                        if (copa_interrupt_received )
                        {
                        }
                        else
                        {
                                restore_flags(flags);
                                interruptible_sleep_on(&WaitQ);
                                save_flags(flags);
                                cli();
                        }
                        copa_interrupt_received=0;
                        restore_flags(flags);
                        break;
                case _IO('r', 1):
                        if ((pDoorbellRegister) && (*pProperties & (1<<1)))
                        {
                                *pDoorbellRegister |= (1<<16);
                                *pDoorbellRegister |= (1<<8);
                                if (copaDebug) printk(KERN_DEBUG "copa :" "COPA_IOC_GENERATE_IRQ pDoorbellRegister 0x%x data 0x%08x\n",(int)pDoorbellRegister,*pDoorbellRegister);
                                Properties = *pProperties;
                                Properties |= (1<<2);
                                *pProperties=Properties;
                                if (copaDebug) printk(KERN_DEBUG "copa :" "COPA_IOC_GENERATE_IRQ Properties 0x%x\n",*pProperties);
                        }
                        else
                                return -EINVAL;
                        break;
                case _IO('r', 2):
                        if ((pDoorbellRegister) && (*pProperties & (1<<1)))
                        {
                                Properties = *pProperties;
                                Properties |= (1<<2);
                                *pProperties=Properties;
                                if (copaDebug) printk(KERN_DEBUG "copa :" "COPA_IOC_ENABLE_RECEIVE_IRQ Properties 0x%x\n",*pProperties);
                                *pDoorbellRegister |= (1<<16);
                                *pDoorbellRegister |= (1<<0);
                                if (copaDebug) printk(KERN_DEBUG "copa :" "COPA_IOC_ENABLE_RECEIVE_IRQ pDoorbellRegister 0x%x data 0x%08x\n",(int)pDoorbellRegister,*pDoorbellRegister);
                        }
                        else
                                return -EINVAL;
                        break;
                case _IO('r', 3):
                        if ((pDoorbellRegister) && (*pProperties & (1<<1)))
                        {
                                Properties = *pProperties;
                                Properties &= ~(1<<2);
                                *pProperties=Properties;
                                if (copaDebug) printk(KERN_DEBUG "copa :" "COPA_IOC_DISABLE_RECEIVE_IRQ Properties 0x%x\n",*pProperties);
                                *pDoorbellRegister &= ~(1<<16);
                                *pDoorbellRegister &= ~(1<<0);
                                if (copaDebug) printk(KERN_DEBUG "copa :" "COPA_IOC_DISABLE_RECEIVE_IRQ pDoorbellRegister 0x%x data 0x%08x\n",(int)pDoorbellRegister,*pDoorbellRegister);
                        }
                        else
                                return -EINVAL;
                        break;
                case _IO('r', 4):
                        if (pDoorbellRegister)
                        {
                                *pDoorbellRegister |= (1<<24);
                                if (copaDebug) printk(KERN_DEBUG "copa :" "COPA_IOC_RESET_RECEIVE_IRQ pDoorbellRegister 0x%x data 0x%08x\n",(int)pDoorbellRegister,*pDoorbellRegister);
                        }
                        else
                                return -EINVAL;
                        break;
                default:
                        if (copaDebug) printk(KERN_DEBUG "copa :" " ioctl-cmd=%x not valid\n",cmd);
                        return -EINVAL;
        }
        return SUCCESS;
}
static struct file_operations copa_fops =
{
    .ioctl = copa_ioctl,
    .mmap = copa_mmap,
    .open = copa_open,
    .release = copa_release
};
int init_module(void)
{
        int result;
    if (! pci_present()) {
        if (copaDebug) printk(KERN_DEBUG "copa :" "init_module: pcibios not present\n");
        return -ENODEV;
    }
    copaf1Dev = pci_find_device(SNI_VENDOR_ID, COPA_F1_DEVICE_ID,copaf1Dev);
    if (copaf1Dev == NULL)
        {
        if (copaDebug) printk(KERN_DEBUG "copa :" "init_module: copernicusAgilent not present!\n");
                return -ENODEV;
    }
        copaRomStart = pci_resource_start(copaf1Dev, 2) & 0xfffffff0;
        pDoorbellAddress = pci_resource_start(copaf1Dev, 0) & 0xfffffff0;
        pvirtRomStart=ioremap(pDoorbellAddress,4*1024);
        if (pvirtRomStart)
        {
                pDoorbellRegister = (unsigned int *)(pvirtRomStart + 0x1C);
                if (copaDebug) printk(KERN_DEBUG "copa :" "pDoorbellRegister 0x%x data 0x%08x\n",(int)pDoorbellRegister,*pDoorbellRegister);
        }
        pvirtRomStart=ioremap_nocache(copaRomStart,4*1024);
        if (pvirtRomStart)
        {
                pProperties = (unsigned int *)(pvirtRomStart + 22*4);
                if (copaDebug) printk(KERN_DEBUG "copa :" "pProperties 0x%x data 0x%x\n",(int)pProperties,*pProperties);
                Properties=*pProperties;
                Properties |= 0x4;
                *pProperties = Properties;
                if (copaDebug) printk(KERN_DEBUG "copa :" "pProperties 0x%x data 0x%x\n",(int)pProperties,*pProperties);
        }
    if (copaDebug) printk(KERN_DEBUG "copa :" "Base Register 0: 0x%x\n",(int)pci_resource_start(copaf1Dev, 0));
    if (copaDebug) printk(KERN_DEBUG "copa :" "Base Register 1: 0x%x\n",(int)pci_resource_start(copaf1Dev, 1));
    if (copaDebug) printk(KERN_DEBUG "copa :" "Base Register 2: 0x%x\n",(int)pci_resource_start(copaf1Dev, 2));
        if (copaDebug) printk(KERN_DEBUG "copa :" "Copernicus(Agilent) found connected to IRQ %d\n",copaf1Dev->irq);
        if (copaf1Dev->irq != 0)
        {
                result=request_irq(copaf1Dev->irq,copa_irq_handler,SA_SHIRQ | SA_INTERRUPT,COPA_DEV,copaf1Dev);
                if (result < 0)
                {
                        if (copaDebug) printk(KERN_DEBUG "copa :" "Bad irq number or handler\n");
                        return result;
                }
        }
    if((copa_major=register_chrdev(copa_major, COPA_DEV, &copa_fops)) < 0) {
        printk(KERN_ERR "copa: unable to get major %d\n", copa_major);
        return -EIO;
    }
    return SUCCESS;
}
void cleanup_module(void)
{
    int err;
    wake_up(&WaitQ);
        if (copaf1Dev->irq != 0)
                free_irq(copaf1Dev->irq,copaf1Dev);
    err = unregister_chrdev(copa_major, COPA_DEV);
    if (err < 0) {
        printk(KERN_ERR "copa: unregister_chrdev(%i) failed!\n", copa_major);
    }
    copaf1Dev = NULL;
    copaf0Dev = NULL;
}
